#!/usr/bin/env python3
import json
import os
import socket
import subprocess

import simplejson as json
import vppapi_if.vppapi_if as vppif
from configobj import ConfigObj
from sysrepo_api import sysrepo_api3 as api

K8S_ENABLED_FILE = '/config/k8s-enabled'
K8S_DISABLED_FILE = '/config/k8s-disabled'
NATIVE_ENABLED_FILE = '/config/native-enabled'
NATIVE_DISABLED_FILE = '/config/native-disabled'
CBOX_FILE = '/config/central-box'
K8S_SINGLE_MACHINE = '/config/k8s-single-machine'
VERIFIED_LICENSE = '/config/license/ucpe_license.lic'


try:
    with open(VERIFIED_LICENSE, "r", encoding='utf-8') as lic_file:
        lics = json.load(lic_file)
        license_features = lics['license']['features']
except:
    license_features = []


native_apps_allowed_by_license = 'native_app' in license_features
k8s_allowed_by_license = 'k8s' in license_features


def checkPidRunning(pidfile):
    """Check For the existence of a unix pid.
    """
    if not os.path.isfile(pidfile):
        return False

    try:
        pid = open(pidfile, "r").readline().strip()
        pid = int(pid)
    except:
        return False

    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True
    return False


class SysrepoSystem(api.SysrepoAPI):
    CONST_DISTR = 'nfvtimeos'
    CONST_MODS = ['main']

    def __init__(self,
                 app_name="system",
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 indexes={"config-server": "server"},
                 namespace="/ucpe:config/system:system",
                 status_namespace="/ucpe:status/system:system",
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            status_namespace=status_namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority)
        self.will_reboot = False

    def _remove_apt_source(self, url):
        self.logger.debug('_remove_apt_source(%s)' % url)
        sr = aptsources.sourceslist.SourcesList()

        to_remove = []
        for sl in sr:
            if sl.disabled or sl.invalid:
                continue

            if sl.dist == self.CONST_DISTR and url == sl.uri:
                to_remove.append(sl)

        if len(to_remove) > 0:
            for rem in to_remove:
                sr.remove(rem)

            sr.save()

    def _apply_hostname(self, hostname=""):
        self.logger.debug('_apply_hostname( %s )' % hostname)
        f = open('/etc/hostname', 'r')
        oldhost = f.readline().rstrip('\n')
        f.close()
        f = open('/etc/hostname', 'w')
        f.write(hostname)
        f.flush()
        os.fsync(f.fileno())
        f.close()

        subprocess.check_output(['hostnamectl', 'set-hostname', hostname])

        if not os.path.isfile('/config/cluster.json'):
            f = open('/opt/vpn-name', 'w')
            f.writelines([hostname + '\n', 'nopass'])
            f.flush()
            os.fsync(f.fileno())
            f.close()

        interfaces = self.get_interfaces()
        for ifname in interfaces:
            if interfaces[ifname].get('type') == 'openvpn' and interfaces[ifname].get('enable', False):
                subprocess.call(
                    'systemctl is-active openvpn@client 2>/dev/null && systemctl --no-block restart openvpn@client.service 2>/dev/null || true',
                    shell=True)
                break

        if os.path.isfile('/config/cluster.json'):
            if self.is_service_active('collectd'):
                subprocess.call('systemctl --no-block restart collectd.service || true', shell=True)
                #subprocess.call('systemctl --no-block restart collectdToLogstash.service || true', shell=True)

        else:
            subprocess.call('systemctl --no-block restart collectd.service || true', shell=True)
            #subprocess.call('systemctl --no-block restart collectdToLogstash.service || true', shell=True)

        subprocess.call('systemctl --no-block restart rsyslog.service || true', shell=True)

        f = open('/etc/hosts', 'r')
        content = f.read()
        content = content.replace(oldhost, hostname)
        f.close()
        f = open('/etc/hosts', 'w')
        f.write(content)
        f.flush()
        os.fsync(f.fileno())
        f.close()

    def _apply_homecalling_timeout(self, timeout=None):
        self.logger.debug('_apply_homecalling_timeout( %s )' % timeout)
        cfg_obj = ConfigObj("/lib/systemd/system/homecalling.service")
        cfg_obj['Service']['TimeoutSec'] = timeout
        self.write_file_to_disk(cfg_obj)
        subprocess.call('systemctl daemon-reload', shell=True)

    def _get_hwtype(self):
        hw_type_file = "/config/.hwtype"
        if not os.path.exists(hw_type_file):
            return ""

        f_hwtype = open(hw_type_file, "r")
        hwtype = f_hwtype.readline().strip()
        f_hwtype.close()
        return hwtype

    def _get_version(self):
        os_version_file = "/etc/nfvtime-os.release"
        if not os.path.exists(os_version_file):
            return ""

        f_version = open(os_version_file, "r")
        version = f_version.readline().strip()
        f_version.close()
        return version

    def _k8s_capabilities_disabled(self):
        if os.path.exists(CBOX_FILE):
            self.logger.debug('Controller ignores worker-kubernetes capabilities')
            return

        self.logger.debug('Disable kubernetes capabilities')
        is_changed = None
        if os.path.exists(K8S_ENABLED_FILE):
            is_changed = True

        try:
            with open(K8S_DISABLED_FILE, 'w'):
                pass

            os.unlink(K8S_ENABLED_FILE)
        except Exception as e:
            self.logger.debug('Disable k8s exception ignore {}'.format(e))

        if is_changed:
            self.execute_shell_command('systemctl --no-block disable kubelet.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block disable containerd.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block disable nfvtime-kubernetes-images.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block disable nfvtime-kubernetes-init.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block disable kube-proxy.service 2>/dev/null || true')
            self.logger.debug('Before call sr-kubenetes-commands.py')
            subprocess.Popen(['/scripts/sr-kubernetes-commands.py', 'k8sdisabled'])
            self.logger.debug('After call sr-kubenetes-commands.py')
            self.execute_shell_command('systemctl --no-block disable sr-kubernetes.service 2>/dev/null || true')

    def _k8s_capabilities_enabled(self):
        if os.path.exists(CBOX_FILE):
            self.logger.debug('Controller ignores worker-kubernetes capabilities')
            return

        # if license is not allowed, we will silently return.
        if not k8s_allowed_by_license:
            self.logger.error('K8S not allowed by a valid license.')
            return

        self.logger.debug('Enable kubernetes capabilities')
        is_changed = None
        if os.path.exists(K8S_DISABLED_FILE):
            is_changed = True

        try:
            state = subprocess.check_output(['systemctl', 'is-enabled', 'sr-kubernetes.service' ])
            if state.decode().rstrip() == 'masked':
                is_changed = True
        except:
            is_changed = True

        try:
            with open(K8S_ENABLED_FILE, 'w'):
                pass

            os.unlink(K8S_DISABLED_FILE)
        except Exception as e:
            self.logger.debug('Enable k8s exception ignore - {}'.format(e))

        if is_changed:
            self.execute_shell_command('systemctl --no-block enable kubelet.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block enable containerd.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block enable nfvtime-kubernetes-images.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block enable nfvtime-kubernetes-init.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block unmask sr-kubernetes.service 2>/dev/null || true')
            self.execute_shell_command('systemctl --no-block enable sr-kubernetes.service 2>/dev/null || true')
            self.logger.info('k8s enabled - need restart')
            self._system_reboot()

    def _k8s_capabilities_unknown(self):
        self.logger.debug('Missing kubernetes capabilities')
        try:
            os.unlink(K8S_ENABLED_FILE)
            os.unlink(K8S_DISABLED_FILE)
        except Exception as e:
            self.logger.debug('Mising kubernetes capabilities {}'.format(e))

    def _native_capabilities_disabled(self):
        self.logger.debug('Disable native capabilities')
        is_changed = None
        if os.path.exists(NATIVE_ENABLED_FILE):
            is_changed = True

        try:
            with open(NATIVE_DISABLED_FILE, 'w'):
                pass

            os.unlink(NATIVE_ENABLED_FILE)
        except Exception as e:
            self.logger.debug('Disable native exception ignore {}'.format(e))

        if is_changed:
            # disable apps, if enabled
            os.system("cd /root/apps/; for dirname in `ls */ -d | cut -f1 -d'/'`; do systemctl disable $dirname.service; done")
            self.logger.info('native disabled - need restart')
            self.will_reboot = True

    def _native_capabilities_enabled(self):
        self.logger.debug('Enable native capabilities')
        is_changed = None

        if os.path.exists(NATIVE_DISABLED_FILE):
            is_changed = True

        try:
            with open(NATIVE_ENABLED_FILE, 'w'):
                pass

            os.unlink(NATIVE_DISABLED_FILE)
        except Exception as e:
            self.logger.debug('Enable native exception ignore - {}'.format(e))

        # if license is not allowed, we will silently return. App start/onboard will not succeed, there is check for license there too.
        if not native_apps_allowed_by_license:
            self.logger.error('Native applications not allowed by a valid license.')
            return

        if is_changed:
            self.logger.info('Native enabled - need restart')
            self.will_reboot = True

    def _native_capabilities_unknown(self):
        self.logger.debug('Missing native capabilities')
        try:
            os.unlink(NATIVE_ENABLED_FILE)
            os.unlink(NATIVE_DISABLED_FILE)
        except Exception as e:
            self.logger.debug('Mising native capabilities {}'.format(e))

    def apply_startup(self):
        self.logger.debug('Apply STARTUP config for system')
        sys_conf = self.get_startup_config(full_config=False, full_path=False)
        managers = self.get_config_servers()
        manager = ""
        for i in managers:
            manager = i
            if manager is not None:
                break

        self.logger.debug('Startup config : %s' % json.dumps(sys_conf))

        if 'config-server' in sys_conf and len(sys_conf['config-server']) > 0:
            self.set_collectd_config_server(manager)
            self.set_kube_proxy(manager, True)
            self.__config_vpp_ippass('tcp', sys_conf['config-server'], 30378)
            self.__config_vpp_ippass('tcp', sys_conf['config-server'], 30636)

        if 'disabled' in sys_conf.get('capabilities').get('worker-kubernetes'):
            self._k8s_capabilities_disabled()
        elif 'enabled' in sys_conf.get('capabilities').get('worker-kubernetes'):
            self._k8s_capabilities_enabled()
        else:
            self._k8s_capabilities_unknown()

        if 'disabled' in sys_conf.get('capabilities').get('native-applications'):
            self._native_capabilities_disabled()
        elif 'enabled' in sys_conf.get('capabilities').get('native-applications'):
            self._native_capabilities_enabled()
        else:
            self._native_capabilities_unknown()

        if 'hostname' in sys_conf:
            self.logger.debug('Apply hostname')
            self._apply_hostname(sys_conf['hostname'])

        if 'timeout' in sys_conf:
            self.logger.debug('Apply HC timeout')
            self._apply_homecalling_timeout(sys_conf['timeout'])

        if 'network-mode' in sys_conf:
            subprocess.call("python3 /opt/bin/switch_ovs_vpp.py --enable {} --reboot".format(sys_conf['network-mode']),
                            shell=True)

        return True

    # Post apply actions callback
    def post_apply_config(self, change_set):
        self.logger.debug('Apply RUNNING config for system')

        for change in change_set:
            if not change["xpath"].startswith(self.namespace):
                continue
            if not 'operation' in change:
                self.logger.error('Malformed change set')
                continue

            if change['operation'] == 'MODIFIED':
                if "network-mode" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    self._system_reboot()
                if "hostname" in change["xpath"]:
                    if os.path.exists('/config/central-box') or os.path.exists('/config/k8s-single-machine'):
                        self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                        self.execute_shell_command("rm -rf /root/.kube > /dev/null 2>&1")
                        self.execute_shell_command("truncate -s 0 /kubernetes/calico-ready > /dev/null 2>&1")
                        self.execute_shell_command('kubeadm reset --cert-dir="/etc/kubernetes/pki" --ignore-preflight-errors=all --force > /dev/null 2>&1')
                        self.execute_shell_command("rm -rf /etc/cni/ /etc/kubernetes/ /root/.kube/ /home/tveapi/.kube/ /var/lib/kubestore/* /var/lib/etcd/* > /dev/null 2>&1")
                        self.execute_shell_command("rm -rf /var/etcd/contiv-data > /dev/null 2>&1")
                        self.execute_shell_command("systemctl enable nfvtime-kubernetes-init.service")
                        self.will_reboot = True
        if self.will_reboot:
            self._system_reboot()

        return

    def apply_running(self, change_set):
        self.logger.debug('Apply RUNNING config for system')
        self.will_reboot = False

        for change in change_set:
            if not change["xpath"].startswith(self.namespace) and not change["xpath"].startswith("/ucpe:config/kubernetes:kubernetes"):
                continue
            if not 'operation' in change:
                self.logger.error('Malformed change set')
                continue

            self.logger.debug('Operation -->> %s' % change['operation'])
            self.logger.debug('XPath -->> %s' % change['xpath'])

            if change['operation'] == 'CREATED':
                self.logger.debug('Created - %s' % change['xpath'])
                if change["xpath"].find("system/config-server") != -1 and change["xpath"].find("/server") != -1:
                    self.set_collectd_config_server(change['new_value'])
                    self.set_kube_proxy(change['new_value'], True)
                    self.__config_vpp_ippass('tcp', change['new_value'], 30378)
                    self.__config_vpp_ippass('tcp', change['new_value'], 30636)

            elif change['operation'] == 'DELETED':
                if change["xpath"].find("system/config-server") != -1 and change["xpath"].find("/server") != -1:
                    self.set_kube_proxy(change['old_value'], False)
                    self.__config_vpp_ippass('tcp', change['old_value'], 30378, add_server=0)
                    self.__config_vpp_ippass('tcp', change['old_value'], 30636, add_server=0)

            elif change['operation'] == 'MODIFIED':
                if "hostname" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    self._apply_hostname(change['new_value'])

                if "timeout" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    self._apply_homecalling_timeout(change['new_value'])

                if "network-mode" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    subprocess.call("python3 /opt/bin/switch_ovs_vpp.py --to {}".format(change['new_value']),
                                    shell=True)

                if 'worker-kubernetes' in change['xpath']:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    if 'disabled' in change['new_value']:
                        self._k8s_capabilities_disabled()
                    elif 'enabled' in change['new_value']:
                        self._k8s_capabilities_enabled()
                    else:
                        self._k8s_capabilities_unknown()

                if 'native-applications' in change['xpath']:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    if 'disabled' in change['new_value']:
                        self._native_capabilities_disabled()
                    elif 'enabled' in change['new_value']:
                        self._native_capabilities_enabled()
                    else:
                        self._native_capabilities_unknown()

                if 'cbox/enabled' in change['xpath']:
                    self.logger.debug(
                        'Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    if 'false' in change['new_value']:
                        self.set_kube_proxy(enable=False)


            elif change['operation'] == 'MOVED':
                self.logger.debug('Moved  - %s' % change['xpath'])

            else:
                self.logger.error('Unknown operation \'%s\'' % change['operation'])

        return True

    def apply_config(self, change_set=None, apply_startup=False):
        ret = False
        if apply_startup:
            ret = self.apply_startup()
        else:
            ret = self.apply_running(change_set)

        return ret

    def status_get_cb(self, xpath, private_data):
        data = { "ucpe:status": { "system:system": { "hwtype": "{}".format(self._get_hwtype()),
                                                     "version": "{}".format(self._get_version())
                                                   }
                                }
               }
        return data

    def set_network_mode(self, value):
        self.set_item("{}/network-mode".format(self.namespace), value)

    def get_network_mode(self):
        return self.get_item("{}/network-mode".format(self.namespace))

    def set_homecalling_enabled(self, value=True):
        self.set_item("{}/homecalling/enabled".format(self.namespace), value)

    def get_homecalling_enabled(self):
        return self.get_item("{}/homecalling/enabled".format(self.namespace))

    def set_homecalling_timeout(self, value=600):
        self.set_item("{}/homecalling/timeout".format(self.namespace), value)

    def get_homecalling_timeout(self):
        return self.get_item("{}/homecalling/timeout".format(self.namespace))

    def set_hostname(self, value):
        self.set_item("{}/hostname".format(self.namespace), value)

    def get_hostname(self):
        return self.get_item("{}/hostname".format(self.namespace))

    def unset_hostname(self):
        self.delete_item("{}/hostname".format(self.namespace))

    def remove_config_servers(self, server=None):
        c_srvs = self.get_config_values(xpath="{}/config-server".format(self.namespace))
        if c_srvs is None:
            return

        for srv in c_srvs:
            if server is None or server != srv:
                self.unset_config_server(srv)

    def set_config_server(self, server):
        self.remove_config_servers(server=server)
        self.set_item("{}/config-server[server='{}']".format(self.namespace, server), server)

    def get_config_server(self, server):
        return self.get_item("{}/config-server[server='{}']".format(self.namespace, server))

    def unset_config_server(self, server):
        self.delete_item("{}/config-server[server='{}']".format(self.namespace, server))

    def get_config_servers(self):
        config = self.get_running_config(full_path=False)
        config_indexed = {}
        if 'config-server' not in config:
            return config_indexed
        for server in config['config-server']:
            config_indexed[server['server']] = server
        return config_indexed

    def get_interfaces(self):
        self.namespace = "/ucpe:config/interfaces:interface"
        config_indexed = {}
        for interface in self.get_running_config(full_path=False):
            config_indexed[interface['name']] = interface

        self.namespace = "/ucpe:config/system:system"
        return config_indexed

    def get_interface_param(self, name, param, default=None):
        self.namespace = "/ucpe:config/interfaces:interface"
        try:
            for interface in self.get_running_config(full_path=False):
                if interface['name'] == name and param in interface:
                    return self.get_item("{}[name='{}']/{}".format("/ucpe:config/interfaces:interface", name, param))
        except:
            return default
        finally:
            self.namespace = "/ucpe:config/system:system"

    def __config_vpp_ippass(self, protocol, server, dst_port, add_server=1):
        if self.get_network_mode() != 'vpp':
            return
        if not checkPidRunning('/run/vpp.pid'):
            return

        if_cfg = self.get_interfaces()
        ippass_ifs = list(filter(lambda x: True
        if (x.get('type', '') == 'ethernet'
            or x.get('type', '') == 'lte'
            or x.get('type', '') == 'wifi')
           or x.get('type', '') == 'pppoe'
           and x.get('ip-passthrough', False)
        else False, list(if_cfg.values())))
        if not ippass_ifs:
            return

        self.logger.debug('__config_vpp_ippass({},{},{}, direction out)'.
                          format(protocol, server, dst_port))

        if str(server).find("http://") > -1:
            try:
                # cut "http://"
                server = server.get("url", None)[7:]
            except AttributeError:
                server = server[7:]

        else:
            if type(server) != str:
                server = str(server)[13:][:-3]

        # resolve hostname
        try:
            server = socket.gethostbyname(server)
        except socket.gaierror:
            # TODO
            self.logger.info("Fail to resolve hostname for {}".format(server))
            return

        self.__set_ippass_enable_dst_port(ippass_ifs, protocol, server, dst_port,
                                          add_server)

    def __set_ippass_enable_dst_port(self,
                                     ippass_ifs, protocol, ip, dst_port,
                                     add, direction=1):
        vpp_if = vppif.VppApiInterface('sr-system')

        for ippass_if in ippass_ifs:
            ippass_host_idx = vpp_if.search_vppif_by_hostifname(ippass_if['name'])
            ippass_vpp_tap_idx = vpp_if.ippass_get_host_if_idx(ippass_host_idx[1])
            if not ippass_vpp_tap_idx:
                continue

            vpp_if.add_del_ippass_permit_host_entry(ippass_vpp_tap_idx, protocol,
                                                    ip=ip, dst_port=dst_port,
                                                    dir_out=direction,
                                                    add=add)

    def set_collectd_config_server(self, manager=None):
        if manager and os.path.exists('/etc/collectd/collectd.conf'):
            #self.execute_shell_command('sed -i "s/tcp:.*/tcp:{}:30826/g" /lib/systemd/system/collectdToLogstash.service'.format(manager))
            self.execute_shell_command('sed -i "s/<Server \\".*\\" \\".*\\"> \\# LOGSTASH_SERVER_LINE/<Server \\"{}\\" \\"30826\\"> \\# LOGSTASH_SERVER_LINE/g" /etc/collectd/collectd.conf'.format(manager))

            if os.path.isfile('/config/cluster.json'):
                if self.is_service_active('collectd'):
                    self.execute_shell_command('systemctl --no-block restart collectd 2>/dev/null || true')
                    #self.execute_shell_command('systemctl --no-block restart collectdToLogstash 2>/dev/null || true')
            else:
                self.execute_shell_command('systemctl --no-block restart collectd 2>/dev/null || true')
                #self.execute_shell_command('systemctl daemon-reload 2>/dev/null || true')
                #self.execute_shell_command('systemctl --no-block restart collectdToLogstash 2>/dev/null || true')

    def set_kube_proxy(self, manager=None, enable=False):
        if os.path.exists(CBOX_FILE) or os.path.exists(K8S_SINGLE_MACHINE):
            if manager and enable:
                self.logger.debug('Enable Kubernetes kubectl proxy on port 31002')
                self.execute_shell_command('systemctl enable kube-proxy.service 2>/dev/null || true')
                self.execute_shell_command('systemctl --no-block start kube-proxy.service 2>/dev/null || true')
                self.execute_shell_command('/sbin/iptables -F kubectl-proxy || true')
                self.execute_shell_command('/sbin/iptables -N kubectl-proxy || true')
                self.execute_shell_command('/sbin/iptables -I INPUT -m tcp -p tcp --dport 31002 -j kubectl-proxy')
                self.execute_shell_command('/sbin/iptables -A kubectl-proxy -j DROP || true')
                # vstojkov: Workaround for NFV-18104- allow incoming traffic from entire X.X.X.0/24 subnet. Maybe security risky.
                self.execute_shell_command('/sbin/iptables -I kubectl-proxy --src {}/24 -j ACCEPT || true'.format(manager))
        if not enable:
            self.logger.debug('Disable Kubernetes kubectl proxy on port 31002')
            self.execute_shell_command('systemctl stop kube-proxy.service 2>/dev/null || true')
            self.execute_shell_command('systemctl disable kube-proxy.service 2>/dev/null || true')
            self.execute_shell_command('/sbin/iptables -F kubectl-proxy || true')

